from numpy import average
from torch import nn
import numpy as np
from array import array

__all__ = ['calIoU']

def calIoU(prediction, label):
    assert prediction.shape == label.shape
    prediction, label = nn.Flatten()(prediction), nn.Flatten()(label)
    res = []
    for i in range(label.shape[0]):
        count = array('i', [0, 0, 0, 0])
        P, L = np.array(prediction[i].cpu()), np.array(label[i].cpu())
        for j in range(len(P)):
            datatype = P[j] * 2 + L[j]
            try:
                count[datatype] += 1
            except IndexError:
                raise ValueError('the value of two tensors should be 0 or 1')
        TN, FN, FP, TP = count
        try:
            IoU = TP / (TP + FP + FN)
        except ZeroDivisionError:
            IoU = 1
        res.append(IoU)
    return average(res)



import torchvision.transforms.functional as F
from functools import partial
from PIL import Image


class Resize(object):
    """Resize the input PIL Images to the given size.

    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (h, w), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``
    """

    def __init__(self, size, interpolation=Image.BILINEAR):
        assert isinstance(size, int) or (isinstance(size, (list, tuple)) and len(size) == 2)
        self.size = size
        self.interpolation = interpolation

    def __call__(self, imgs):
        """
        Args:
            imgs (PIL Image): Images to be scaled.

        Returns:
            PIL Image: Rescaled image.
        """
        resize = partial(F.resize(size=self.size, interpolation=self.interpolation))
        if not isinstance(imgs, (list, tuple)):
            return resize(imgs)
        return list(map(resize, imgs))

    def __repr__(self):
        return self.__class__.__name__ + "(size={0})".format(self.size)



class Compose(object):
    """Composes several transforms together.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.

    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, imgs):
        for t in self.transforms:
            imgs = t(imgs)
        return imgs

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string



import torchvision.transforms.functional as F


class ToTensor(object):
    """Convert a ``PIL Image`` or ``numpy.ndarray`` to tensor.

    Converts a PIL Image or numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.FloatTensor of shape (C x H x W) in the range [0.0, 1.0].
    """

    def __call__(self, pics):
        """
        Args:
            pic (PIL Image or numpy.ndarray): Image to be converted to tensor.

        Returns:
            Tensor: Converted image.
        """
        if not isinstance(pics, (list, tuple)):
            return F.to_tensor(pics)
        return list(map(F.to_tensor, pics))

    def __repr__(self):
        return self.__class__.__name__ + "()"



import numpy as np
from torch.utils.data import Dataset
import os
from PIL import Image
from .functional import pick
from torchvision import transforms
from . import Cotrans

class IIset(Dataset):
    def __init__(self, raw_root, label_root, input_shape=(480, 320), output_shape=(480, 320)):
        self.raw_root, self.label_root = raw_root, label_root
        self.input_shape, self.output_shape = input_shape, output_shape
        self.namelist = os.listdir(raw_root)

    def __getitem__(self, i):
        name = self.namelist[i]
        imgIn = Image.open(os.path.join(self.raw_root, name))
        try:
            imgOut = Image.open(os.path.join(self.label_root, name))
        except FileNotFoundError:
            raise FileNotFoundError(f'Image<{name}> has no corresponding label')
        assert imgIn.size == imgOut.size
        
        imgOut = Image.fromarray(np.vectorize(lambda x: 255 if x == 255 else 0)(np.asarray(imgOut)[..., 0]).astype(np.uint8))
        cotrans = Cotrans.Compose([
            Cotrans.RandomResizedCrop(self.input_shape),
            Cotrans.RandomHorizontalFlip()
        ])
        while True:
            imgInCrop, imgOutCrop = cotrans([imgIn, imgOut])
            imgOutCrop = transforms.CenterCrop(self.output_shape)(imgOutCrop)
            if pick(np.average(imgOutCrop) / 255, 0.1):
                imgIn, imgOut = imgInCrop, imgOutCrop
                break
        imgIn, imgOut = Cotrans.ToTensor()([imgIn, imgOut])
        imgIn = transforms.Compose([
            transforms.ColorJitter(0.5, 0.5, 0.5, 0.5),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        ])(imgIn)
        return imgIn.float(), imgOut.squeeze().long()
    

    def __len__(self):
        return len(self.namelist)


if __name__ == "__main__":
    pass



import random
import torch
from torch.utils.data.dataset import random_split
from torchvision.transforms import ToPILImage
import numpy as np
__ALL__ = ['pick', 'ratio_split', 'rev_trans']

__all__ = __ALL__ + ['__ALL__']


def pick(probability, base=0):
    compare = random.random()
    return True if probability + base > compare else False


def ratio_split(dataset, percentages):
    if round(sum(percentages), 3) != 1.0:
        raise ValueError(
            'Sum of input percentages({}) does not equal 1'.format(sum(percentages)))
    lenList = list(map(lambda x: int(x * len(dataset)), percentages))
    lenList[-1] = len(dataset) - sum(lenList[0:-1])
    return random_split(dataset, lenList)


def rev_trans(tensor, mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]):
    mean = torch.FloatTensor(mean).view(3,1,1).expand(*tensor.shape)
    std = torch.FloatTensor(std).view(3,1,1).expand(*tensor.shape)
    return ToPILImage()(tensor * std + mean)




import copy
from tqdm import tqdm
import torch
import pandas as pd
import torch.nn.functional as F
from .functional import calIoU

def test(model, criterion, valLoader):

    device = next(model.parameters()).device
    val_loss, val_IoU, batchs = 0.0, 0.0, 0
    model.eval() 
    for X, Y in tqdm(valLoader):
        X, Y = X.to(device), Y.to(device)
        out = model(X)
        loss = criterion(out, Y)
        batchs += 1
        val_loss += loss.item()
        prediction = torch.argmax(out, 1)
        val_IoU += calIoU(prediction, Y)
    return {'loss': val_loss / batchs, 'IoU': val_IoU / batchs}



import random
import torch
import torchvision
import torchvision.transforms.functional as F
import numbers
import numpy as np
from PIL import Image
import math

class RandomCrop(object):
    def __init__(self, size, padding=0):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.padding = padding

    @staticmethod
    def get_params(img, output_size):
        w, h = img.size
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def __call__(self, imgs):
        i, j, h, w = self.get_params(imgs[0], self.size)

        def crop(img):
            if isinstance(img, torch.Tensor):
                return torchvision.transforms.ToTensor()(crop(torchvision.transforms.ToPILImage()(img)))
            if isinstance(img, np.ndarray):
                return np.asarray(crop(Image.fromarray(img.astype(np.uint8))))
            if self.padding > 0:
                img = F.pad(img, self.padding)
            return F.crop(img, i, j, h, w)

        if not isinstance(imgs, (list, tuple)):
            return crop(imgs)
        return list(map(crop, imgs))

    def __repr__(self):
        return self.__class__.__name__ + f"(size={self.size})"

class RandomResizedCrop(object):
    def __init__(
            self,
            size,
            scale=(0.08, 1.0),
            ratio=(3.0 / 4.0, 4.0 / 3.0),
            interpolation=Image.BILINEAR,
    ):
        self.size = (size, size) if isinstance(size, int) else size
        self.interpolation = interpolation
        self.scale = scale
        self.ratio = ratio

    @staticmethod
    def get_params(img, scale, ratio):
        for attempt in range(10):
            area = img.size[0] * img.size[1]
            target_area = random.uniform(*scale) * area
            aspect_ratio = random.uniform(*ratio)

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if random.random() < 0.5:
                w, h = h, w

            if w <= img.size[0] and h <= img.size[1]:
                i = random.randint(0, img.size[1] - h)
                j = random.randint(0, img.size[0] - w)
                return i, j, h, w

        # Fallback
        w = min(img.size[0], img.size[1])
        i = (img.size[1] - w) // 2
        j = (img.size[0] - w) // 2
        return i, j, w, w

    def __call__(self, imgs):
        i, j, h, w = self.get_params(imgs[0], self.scale, self.ratio)

        def resized_crop(img):
            if isinstance(img, torch.Tensor):
                return torchvision.transforms.ToTensor()(resized_crop(torchvision.transforms.ToPILImage()(img)))
            if isinstance(img, np.ndarray):
                return np.asarray(resized_crop(Image.fromarray(img.astype(np.uint8))))
            return F.resized_crop(img, i, j, h, w, self.size, self.interpolation)

        return list(map(resized_crop, imgs)) if isinstance(imgs, (list, tuple)) else resized_crop(imgs)

    def __repr__(self):
        return self.__class__.__name__ + "(size={0})".format(self.size)



# -*- coding:utf-8 -*-

from . import dataset, models, train, functional, optim

__author__ = 'Jewelry'

__all__ = ['models', 'dataset', 'train', 'functional', 'optim']

# 12.2
# extend plot_procedure to dataframes

# 12.1
# add plot_procedure grid_resize eval_tranforms fuse

# 11.30
# verify the centerCrop and oversampling of IIset
# add plot_pair 
# delete nn temporarily

# 11.29
# add rev_trans
# debug IIset
# add non-padding unet
# add choice 'centercrop' to IIset
# rectify plot4 -> plots

# 11.28
# add optim and import optimizers from torch_optimizer
# add Swish 
# RandomCrop -> RandomResizedCrop
# TODO upload to pypi

# 11.23
# 1. add train_unit test_unit
# 2. unet rectified
# 3. calIoU fix the issue of nan
# 4. Important: transforms->Cotrans, Never mix torch stardart transforms and Cotrans up!!



import torch
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from .dataset.functional import pick, ratio_split, rev_trans
from torchvision import transforms
__all__ = ['plots', 'plot_pair', 'plot_procedure', 'grid_resize', 'eval_tranforms', 'fuse']
__all__ += ['pick', 'ratio_split', 'rev_trans']


def plots(figs, figsize, layout, save=''):
    plt.figure(figsize=figsize)
    assert len(layout) == 2
    for i in range(layout[0] * layout[1]):
        plt.subplot(*layout, i + 1)
        if isinstance(figs[i], (np.ndarray, torch.Tensor)) and len(figs[i].shape) == 2:
            plt.imshow(figs[i], 'Greys_r')
            continue
        plt.imshow(figs[i])
    if save:
        plt.savefig(save, facecolor='white')
    plt.show()


def plot_pair(sample, figsize=(10, 6)):
    plots([rev_trans(sample[0]), sample[1].float()], figsize=figsize, layout=(1, 2))


def plot_procedure(df, keys):
    def _plot(df, keys):
        if isinstance(df, str):
            df = pd.read_excel(df)
        epochs = np.arange(1, len(df) + 1)
        for key in keys:
            plt.plot(epochs, df[key], label=key)
    plt.figure(figsize=(8, 5))
    if isinstance(df, (list, tuple)):
        for Slc in df:
            _plot(Slc, keys)
    else:
        _plot(df, keys)
    plt.legend()
    plt.xlabel("epoch")
    plt.show()


def grid_resize(image, grid):
    width, height = image.size
    image = image.resize((width // grid * grid, height // grid * grid))
    return image


def eval_tranforms(image, norm=[[0.485, 0.456, 0.406], [0.229, 0.224, 0.225]]):
    trans = transforms.Compose([transforms.ToTensor(), transforms.Normalize(*norm)])
    return trans(image).unsqueeze(0)


def fuse(raw, label):
    image = np.array(raw) * 0.8
    image[..., 1] = image[..., 1] + 0.2 * np.array(label.squeeze(0).argmax(dim=0).cpu() * 255)
    return image.astype(np.uint8)



import torchvision.transforms.functional as F
from ..functional import pick


class RandomHorizontalFlip(object):
    def __call__(self, imgs):
        if not isinstance(imgs, (list, tuple)):
            return F.hflip(imgs) if pick(0.5) else imgs
        return list(map(F.hflip, imgs)) if pick(0.5) else imgs

    def __repr__(self):
        return self.__class__.__name__ + '()'


class RandomVerticalFlip(object):
    def __call__(self, imgs):
        if not isinstance(imgs, (list, tuple)):
            return F.vflip(imgs) if pick(0.5) else imgs
        return list(map(F.vflip, imgs)) if pick(0.5) else imgs

    def __repr__(self):
        return self.__class__.__name__ + '()'



import torch
from torch import nn
from torchvision.models import vgg19

class FCN8s(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        # num_classes:训练数据的类别
        self.num_classes = num_classes
        model_vgg19 = vgg19(pretrained=True)
        ## 不使用vgg19网络中的后面的AdaptiveAvgPool2d和Linear层
        self.base_model = model_vgg19.features
        ## 定义几个需要的层操作，并且使用转置卷积将特征映射进行升维
        self.relu    = nn.ReLU(inplace=True)
        self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=3, stride=2,
                                          padding=1, dilation=1, output_padding=1)
        self.bn1     = nn.BatchNorm2d(512)
        self.deconv2 = nn.ConvTranspose2d(512, 256, 3, 2, 1, 1, 1)
        self.bn2     = nn.BatchNorm2d(256)
        self.deconv3 = nn.ConvTranspose2d(256, 128, 3, 2, 1, 1, 1)
        self.bn3     = nn.BatchNorm2d(128)
        self.deconv4 = nn.ConvTranspose2d(128, 64, 3, 2, 1, 1, 1)
        self.bn4     = nn.BatchNorm2d(64)
        self.deconv5 = nn.ConvTranspose2d(64, 32, 3, 2, 1, 1, 1)
        self.bn5     = nn.BatchNorm2d(32)
        self.classifier = nn.Conv2d(32, num_classes, kernel_size=1)
    
        ## vgg19中MaxPool2d所在的层
        self.layers = {"4": "maxpool_1","9": "maxpool_2", 
                       "18": "maxpool_3", "27": "maxpool_4", 
                       "36": "maxpool_5"}

    def forward(self, x):
        output = {}
        nametag = 0
        for i, (name, layer) in enumerate(self.base_model._modules.items()):
            ## 从第一层开始获取图像的特征
            x = layer(x)
            ## 如果是layers参数指定的特征，那就保存到output中
            if name in self.layers:
                output[self.layers[name]] = x
        x5 = output["maxpool_5"]  # size=(N, 512, x.H/32, x.W/32)
        x4 = output["maxpool_4"]  # size=(N, 512, x.H/16, x.W/16)
        x3 = output["maxpool_3"]  # size=(N, 256, x.H/8,  x.W/8)
        ## 对特征进行相关的转置卷积操作,逐渐将图像放大到原始图像大小
        # size=(N, 512, x.H/16, x.W/16)
        score = self.relu(self.deconv1(x5))
        # 对应的元素相加, size=(N, 512, x.H/16, x.W/16)
        score = self.bn1(score + x4)  
        # size=(N, 256, x.H/8, x.W/8)
        score = self.relu(self.deconv2(score)) 
        # 对应的元素相加, size=(N, 256, x.H/8, x.W/8)
        score = self.bn2(score + x3)  
        # size=(N, 128, x.H/4, x.W/4)
        score = self.bn3(self.relu(self.deconv3(score))) 
        # size=(N, 64, x.H/2, x.W/2)
        score = self.bn4(self.relu(self.deconv4(score)))
        # size=(N, 32, x.H, x.W)
        score = self.bn5(self.relu(self.deconv5(score)))  
        score = self.classifier(score)
        return score  # size=(N, n_class, x.H/1, x.W/1)



from setuptools import setup, find_packages

with open("README.rst", "r") as f:
    long_description = f.read()
setup(
    name='torchflame',
    packages=find_packages(),
    version='0.1.20201203.1',
    license='MIT',
    description='a small extension of pytorch writen and collected by Jewelry',
    long_description=long_description,
    author='Jewelry',
    author_email='3180105772@zju.edu.cn',
    classifiers=[],
    install_requires=['torch', 'torchvision', 'torch_optimizer', 'numpy', 'matplotlib', 'pandas', 'xlrd'],
    python_requires='>=3.6'
)



import copy
import pandas as pd
import numpy as np
from .train_unit import train
from .test_unit import test
import torch
import os


def train_overall(model, criterion, optimizer, trainLoader, valLoader, num_epochs, saveaddr=''):
    '''
    model needs to move to device you want.
    '''
    best_model_wts = copy.deepcopy(model.state_dict())
    best_loss = np.inf
    train_loss_all, train_IoU_all, val_loss_all, val_IoU_all = [], [], [], []
    print("Train on", next(model.parameters()).device)
    for epoch in range(1, num_epochs + 1):
        print("Epoch {}/{}".format(epoch, num_epochs))
        record = train(model, criterion, optimizer, trainLoader)
        train_loss_all.append(record['loss'])
        train_IoU_all.append(record['IoU'])
        print(f"EPOCH {epoch}: Train Loss: {record['loss']:.4f}, Train IoU: {record['IoU']:.3f}")
        record = test(model, criterion, valLoader)
        val_loss_all.append(record['loss'])
        val_IoU_all.append(record['IoU'])
        print(f"EPOCH {epoch}: Val Loss: {record['loss']:.4f}, Val IoU: {record['IoU']:.3f}")
        if saveaddr:
            torch.save(model, os.path.join(saveaddr, f'unet_{epoch}.pkl'))

        if record['loss'] < best_loss:
            best_loss = record['loss']
            best_model_wts = copy.deepcopy(model.state_dict())

    train_process = pd.DataFrame(
        data={
            "epoch": range(1, num_epochs + 1),
            "train_loss_all": train_loss_all,
            'train_IoU_all': train_IoU_all,
            "val_loss_all": val_loss_all,
            'val_IoU_all': val_IoU_all
        })
    model.load_state_dict(best_model_wts)
    return model, train_process



from .Compose import Compose
from .RandomCrop import RandomCrop, RandomResizedCrop
from .RandomFlip import RandomHorizontalFlip, RandomVerticalFlip
from .Resize import Resize
from .ToTensor import ToTensor

__all__ = ['Compose', 'RandomCrop', 'RandomHorizontalFlip', 'RandomResizedCrop', 'RandomVerticalFlip', 'Resize', 'ToTensor']



from tqdm import tqdm
import torch
from .functional import calIoU


def train(model, criterion, optimizer, trainLoader): 

    train_loss, train_IoU, batchs = 0.0, 0.0, 0
    device = next(model.parameters()).device
    model.train()
    for X, Y in tqdm(trainLoader):
        optimizer.zero_grad()
        X, Y = X.to(device), Y.to(device)
        out = model(X)
        loss = criterion(out, Y)
        loss.backward()
        optimizer.step()
        batchs += 1
        train_loss += loss.item()
        prediction = torch.argmax(out, 1)
        train_IoU += calIoU(prediction, Y)
    return {'loss': train_loss / batchs, 'IoU': train_IoU / batchs}
